@@ -28,7 +28,7 @@ import mimetypes
 import sys
 from functools import partial
 
-from fringe_utils.utils import Helper 
+from fringe_utils.utils import Helper, MimeTypes, GSheetsApiDefinedValues
 
 class GSuite():
     scopes = [
@@ -39,12 +39,16 @@ class GSuite():
         'https://www.googleapis.com/auth/presentations'
     ]
 
-    def __init__(self, secrets_path='../creds/oauth-creds-gsuite-api.json'):
-        dirname = os.path.dirname(__file__)
+    def __init__(self, secrets_path='../creds/oauth-creds-gsuite-api.json', pickle_name='token'):
+        dirname = os.path.dirname(os.path.abspath(__file__))
         self.secretsPath = os.path.join(dirname, secrets_path)
+        self.mimetypes = MimeTypes()
+        self.pickle_name = pickle_name
 
     def connect(self, service='sheets', version='v4', scopes=scopes, 
-        pathToCreds=None, pickleName='token'):
+        pathToCreds=None, pickleName=None):
+        if not pickleName:
+            pickleName = self.pickle_name if self.pickle_name else 'token'
         if pathToCreds is None: pathToCreds = self.secretsPath
 
         creds = None
@@ -69,24 +73,26 @@ class GSuite():
 
 class GSheets():
     apiDocs = 'https://developers.google.com/sheets/api/reference/rest'
-    major_dimension = ['ROWS', 'COLUMNS']
-    value_input_option = ['RAW', 'USER_ENTERED']
-    value_render_option = ['FORMATTED_VALUE', 'UNFORMATTED_VALUE', 'FORMULA']
-    dateTimeRenderOption = ['SERIAL_NUMBER', 'FORMATTED_STRING']
+    enums = GSheetsApiDefinedValues()
+    # major_dimension = ['ROWS', 'COLUMNS']
+    # value_input_option = ['RAW', 'USER_ENTERED']
+    # value_render_option = ['FORMATTED_VALUE', 'UNFORMATTED_VALUE', 'FORMULA']
+    # dateTimeRenderOption = ['SERIAL_NUMBER', 'FORMATTED_STRING']
 
     test_ssid = '1wZwiJrkR2p_VhSa-KYm-O2i35RXCrd9aVkCm4q8IZs4'
     test_get_range = 'Sheet1!A1:D14'
     test_update_range = 'Sheet1!A20'
 
-    def __init__(self, scopes=None, secretsPath=None):
+    def __init__(self, scopes=None, secretsPath=None, pickle_name=None):
         gsuite = partial(GSuite)
         if secretsPath: gsuite = partial(gsuite, secrets_path=secretsPath)
-        gsuite_connect = partial(gsuite().connect)
+        if pickle_name: gsuite = partial(gsuite, pickle_name=pickle_name)
+        gsuite_connect = partial(gsuite().connect, service='sheets', version='v4')
         if scopes: gsuite_connect = partial(gsuite_connect, scopes=scopes)
         self.service = gsuite_connect()
 
     def testGet(self, ssid=test_ssid, range=test_get_range, 
-        valueRenderOption=value_render_option[0]):
+        valueRenderOption=enums.value_render_option.FORMATTED_VALUE):
 
         sheets = self.service.spreadsheets()
         result = sheets.values().get(
@@ -101,7 +107,7 @@ class GSheets():
 
         value_range_body = {
             "range": range,
-            "majorDimension": self.major_dimension[0],
+            "majorDimension": self.enums.major_dimension.ROWS,
             "values": [
                 [1234, 34123],
                 [1234, 17634, 1234]
@@ -115,8 +121,15 @@ class GSheets():
         response = request.execute()
         return response
     
-    def get(self, ssid, range, valueRenderOption=value_render_option[0], 
-        majorDimension=major_dimension[0]):
+    def get(self, ssid, range, valueRenderOption=enums.value_render_option.FORMATTED_VALUE, 
+        majorDimension=enums.major_dimension.ROWS):
+        '''
+        Get values from a spreadsheets with id `ssid` for the specified `range`.
+        - `ssid`: id of the Google Sheets spreadhseet
+        - `range`: alphanumeric range
+        - `valueRenderOption`: use `self.enums.value_render_option` for options
+        - `majorDimensionn`: use `self.enums.major_dimension` for options
+        '''
 
         sheets = self.service.spreadsheets()
         result = sheets.values().get(
@@ -129,9 +142,22 @@ class GSheets():
         return values
     
     def update(self, ssid, _range, valuesDf, 
-        majorDimension=major_dimension[0], valueInputOption=value_input_option[0], 
-        includeValuesInResponse=False, responseValueRenderOption=value_render_option[0], 
-        responseDateTimeRenderOption=dateTimeRenderOption[0], fillnaWithBlank=True):
+        majorDimension=enums.major_dimension.ROWS, valueInputOption=enums.value_input_option.RAW, 
+        includeValuesInResponse=False, responseValueRenderOption=enums.value_render_option.FORMATTED_VALUE, 
+        responseDateTimeRenderOption=enums.date_time_render_option.SERIAL_NUMBER, fillnaWithBlank=True):
+        '''
+        Updates values from a spreadsheets with id `ssid` for the specified `range`.
+        # Parameters:
+        - `ssid` (str): id of the Google Sheets spreadhseet
+        - `_range` (str): alphanumeric range
+        - `valuesDf` (DataFrame): a pandas dataframe of values
+        - `majorDimension`: use `self.enums.major_dimension` for 'options'
+        - `valueRenderOption`: use `self.enums.value_render_option` for options
+        - `includeValuesInResponse` (bool): default `False`
+        - `responseValueRenderOption`: use `enums.value_render_option` for options
+        - `responseDateTimeRenderOption`: use `enums.date_time_render_option` for options
+        - `fillnaWithBlank` (bool): default `True`
+        '''
         
         values = (valuesDf.fillna('') if fillnaWithBlank else valuesDf).values.tolist()
         value_range_body = {
@@ -154,22 +180,68 @@ class GSheets():
     def clear(self, ssid, _range):
         ''
     
-    def batchGet(self, ssid, range, valueRenderOption=value_render_option[0], 
-        majorDimension=major_dimension[0]):
+    def batchGet(self, ssid, range, valueRenderOption=enums.value_render_option.FORMATTED_VALUE, 
+        majorDimension=enums.major_dimension.ROWS):
+        ''
+    
+    def batchGetByDataFilter(self, ssid, range, valueRenderOption=enums.value_render_option.FORMATTED_VALUE, 
+        majorDimension=enums.major_dimension.ROWS):
         ''
     
     def batchUpdate(self, ssid, _ranges, valuesDf, 
-        majorDimension=major_dimension[0], valueInputOption=value_input_option[0], 
-        includeValuesInResponse=False, responseValueRenderOption=value_render_option[0], 
-        responseDateTimeRenderOption=dateTimeRenderOption[0], fillnaWithBlank=True):
+        majorDimension=enums.major_dimension.ROWS, valueInputOption=enums.value_input_option.RAW, 
+        includeValuesInResponse=False, responseValueRenderOption=enums.value_render_option.FORMATTED_VALUE, 
+        responseDateTimeRenderOption=enums.date_time_render_option.SERIAL_NUMBER, fillnaWithBlank=True):
+        ''
+    
+    def batchUpdateByDataFilter(self, ssid, _ranges, valuesDf, 
+        majorDimension=enums.major_dimension.ROWS, valueInputOption=enums.value_input_option.RAW, 
+        includeValuesInResponse=False, responseValueRenderOption=enums.value_render_option.FORMATTED_VALUE, 
+        responseDateTimeRenderOption=enums.date_time_render_option.SERIAL_NUMBER, fillnaWithBlank=True):
         ''
     
     def batchClear(self, ssid, _ranges):
         ''
     
-    def append(self, ssid, _range, valueInputOption=value_input_option[0]):
+    def batchClearByDataFilter(self, ssid, _ranges):
         ''
     
+    def append(self, ssid, _range, valuesDf, 
+        majorDimension=enums.major_dimension.ROWS, valueInputOption=enums.value_input_option.RAW, 
+        includeValuesInResponse=False, responseValueRenderOption=enums.value_render_option.FORMATTED_VALUE, 
+        responseDateTimeRenderOption=enums.date_time_render_option.SERIAL_NUMBER, fillnaWithBlank=True):
+        '''
+        Appends values to spreadsheets with id `ssid` at the end of the specified `range`.
+        # Parameters:
+        - `ssid` (str): id of the Google Sheets spreadhseet
+        - `_range` (str): alphanumeric range
+        - `valuesDf` (DataFrame): a pandas dataframe of values
+        - `majorDimension`: use `self.enums.major_dimension` for 'options'
+        - `valueInputOption`: use `self.enums.value_input_option` for options
+        - `includeValuesInResponse` (bool): default `False`
+        - `responseValueRenderOption`: use `enums.value_render_option` for options
+        - `responseDateTimeRenderOption`: use `enums.date_time_render_option` for options
+        - `fillnaWithBlank` (bool): default `True`
+        '''
+        
+        values = (valuesDf.fillna('') if fillnaWithBlank else valuesDf).values.tolist()
+        value_range_body = {
+            "range": _range,
+            "majorDimension": majorDimension,
+            "values": values
+        }
+        req = self.service.spreadsheets().values().append(
+            spreadsheetId = ssid, 
+            range = _range, 
+            valueInputOption = valueInputOption,
+            body = value_range_body,
+            includeValuesInResponse = includeValuesInResponse,
+            responseValueRenderOption = responseValueRenderOption,
+            responseDateTimeRenderOption = responseDateTimeRenderOption
+        )
+        res = req.execute()
+        return res
+    
     def addCellBorders(self, ssid, _range, sheetId, top, bottom, innerHorizontal):
         ''
 
@@ -179,17 +251,17 @@ class GSheets():
         c = col[-1]
         num = ord(c) - first + 1
         if len(col) > 1:
-            num = num + 26*getColumnNumber(col[:-1])
+            num = num + 26*self.getColumnNumber(col[:-1])
         return num
     
     def getIndexesFromRange(self, _range):
         s = re.search(r'(?i)([a-z]*)([0-9]*)(:([a-z]*)([0-9]*))?', _range)
         c1 = s.group(1)
         if c1:
-            c1 = getColumnNumber(c1)
+            c1 = self.getColumnNumber(c1)
         c2 = s.group(4)
         if c2:
-            c2 = getColumnNumber(c2)
+            c2 = self.getColumnNumber(c2)
         r1 = s.group(2)
         if not r1:
             r1 = 1
@@ -209,9 +281,10 @@ class GSheets():
 
 class GSlides():
 
-    def __init__(self, scopes=None, secretsPath=None):
+    def __init__(self, scopes=None, secretsPath=None, pickle_name=None):
         gsuite = partial(GSuite)
         if secretsPath: gsuite = partial(gsuite, secrets_path=secretsPath)
+        if pickle_name: gsuite = partial(gsuite, pickle_name=pickle_name)
         gsuite_connect = partial(gsuite().connect, service='slides', version='v1')
         if scopes: gsuite_connect = partial(gsuite_connect, scopes=scopes)
         self.service = gsuite_connect()
@@ -233,9 +306,10 @@ class GMail():
     labelListVisibilities = ['labelShowIfUnread', 'labelHide', 'labelShow']
     messageListVisibilities = ['hide', 'show']
 
-    def __init__(self, scopes=None, secretsPath=None):
+    def __init__(self, scopes=None, secretsPath=None, pickle_name=None):
         gsuite = partial(GSuite)
         if secretsPath: gsuite = partial(gsuite, secrets_path=secretsPath)
+        if pickle_name: gsuite = partial(gsuite, pickle_name=pickle_name)
         gsuite_connect = partial(gsuite().connect, service='gmail', version='v1')
         if scopes: gsuite_connect = partial(gsuite_connect, scopes=scopes)
         self.service = gsuite_connect()
@@ -401,9 +475,10 @@ class GMail():
 class GDrive():
     apiDocs = 'https://developers.google.com/drive/api/v3/reference'
 
-    def __init__(self, scopes=None, secretsPath=None):
+    def __init__(self, scopes=None, secretsPath=None, pickle_name=None):
         gsuite = partial(GSuite)
         if secretsPath: gsuite = partial(gsuite, secrets_path=secretsPath)
+        if pickle_name: gsuite = partial(gsuite, pickle_name=pickle_name)
         gsuite_connect = partial(gsuite().connect, service='drive', version='v3')
         if scopes: gsuite_connect = partial(gsuite_connect, scopes=scopes)
         self.service = gsuite_connect()
